#pragma once

#include <cstdint>

namespace icpp_hash_library_utils {

// Utility function to check if we are running on little-endian
bool is_little_endian();

// Utility function to check if we are running on big-endian
bool is_big_endian();

}